package smartresource.core;

import java.io.*;

import java.util.Vector;
import java.util.Calendar;

import org.openrdf.rio.rdfxml.*; 
import org.openrdf.rio.*; 
import org.openrdf.model.*; 

import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;

public class AssignRoleBehavior extends ReusableAtomicBehavior{

   public void action() {
                	
	String script_file=getParameterValue("input");
	String init=getParameterValue("startup");
	if(init==null) init="false";
	String emptyBeliefs=getParameterValue("emptyBeliefs");
	if(emptyBeliefs==null) emptyBeliefs="false";
	String emptyGoals=getParameterValue("emptyGoals");
	if(emptyGoals==null) emptyGoals="false";
	String emptyRules=getParameterValue("emptyRules");
	if(emptyRules==null) emptyRules="false";

	if(emptyBeliefs.equals("true"))
		((SmartResourceAgent)myAgent).beliefs=new Vector();

	if(emptyGoals.equals("true")){
		((SmartResourceAgent)myAgent).goals=new Vector();
		((SmartResourceAgent)myAgent).goals_under_work=new Vector();
	}

	if(emptyRules.equals("true"))
		((SmartResourceAgent)myAgent).activities=new Vector();
	

	
	String curr_role=script_file.substring(script_file.lastIndexOf("/")+1,script_file.lastIndexOf("."));

	RDFStatementHandler myStatementHandler = new RDFStatementHandler();
	RDFErrorHandler myParseErrorListener = new RDFErrorHandler();

	System.setProperty("org.xml.sax.driver","org.apache.xerces.parsers.SAXParser");

	Parser parser = new RdfXmlParser();
	parser.setStatementHandler(myStatementHandler);
	parser.setParseErrorListener(myParseErrorListener);
	parser.setVerifyData(true);
	parser.setStopAtFirstError(false);

        try{	
	        InputStreamReader iread=new InputStreamReader(new FileInputStream(script_file));
		BufferedReader ff=new BufferedReader(iread);

		parser.parse(ff, "http://www.smartresource.com/"+((SmartResourceAgent)myAgent).getLocalName()+"/"+curr_role+"/");
	}catch(Exception e){
		e.printStackTrace(); 

		//to avoid harmful behavior
		((SmartResourceAgent)myAgent).beliefs=new Vector();
        	((SmartResourceAgent)myAgent).goals=new Vector();
		((SmartResourceAgent)myAgent).goals_under_work=new Vector();
		((SmartResourceAgent)myAgent).goals_under_work=new Vector();
		((SmartResourceAgent)myAgent).activities=new Vector();
		success=false;
		return;
	}

	if(!init.equals("true")){
		print(curr_role+" role is assigned");
		((SmartResourceAgent)myAgent).myRoles.add(curr_role);
	}

	DFAgentDescription dfd = new DFAgentDescription();
	dfd.setName(myAgent.getAID());
		
	for(int i=0;i<((SmartResourceAgent)myAgent).myRoles.size();i++){
		ServiceDescription sd = new ServiceDescription();
		sd.setName((String)((SmartResourceAgent)myAgent).myRoles.elementAt(i));
		sd.setType("SmartResource");
		dfd.addServices(sd);
	}

	try{
		DFService.deregister(myAgent);
	}catch (FIPAException fe) {}
	try{
		DFService.register(myAgent, dfd);
	}catch (FIPAException fe) {fe.printStackTrace();}

   }

   class RDFErrorHandler implements ParseErrorListener{
		public void error(String msg, int lineNo, int colNo){
		     System.out.println("error! "+msg);
		} 
		public void fatalError(String msg, int lineNo, int colNo){
		     System.out.println("fatal! "+msg);
		}
                public void warning(String msg, int lineNo, int colNo){
		     System.out.println("warning! "+msg);
                }
    }

   class RDFStatementHandler implements StatementHandler{
                Vector belief_uris;
                Vector goal_uris;
                Vector activities_uris;
                String gui_uri;
                String server_uri;
                String today;

                public RDFStatementHandler(){
                	belief_uris=new Vector();
                	goal_uris=new Vector();
                	activities_uris=new Vector();

			Calendar now=Calendar.getInstance();
			today=String.valueOf(now.get(Calendar.DAY_OF_MONTH))+"."+String.valueOf(now.get(Calendar.MONTH))+"."+String.valueOf(now.get(Calendar.YEAR));
                }

		public void handleStatement(Resource subject, URI predicate, Value object){
			if(subject instanceof URI){
				String subj=((URI)subject).getLocalName();
				String subj_uri=((URI)subject).getURI();
				String pred=predicate.getLocalName();
				String pred_ns=predicate.getNamespace();
				String obj="";
				String obj_ns="";
				if(object instanceof Literal){
					obj=((Literal)object).getLabel();
				}
				else if(object instanceof URI) {
					obj=((URI)object).getLocalName();
					obj_ns=((URI)object).getNamespace();
				}
                                obj=obj.replaceAll("%AgentName%",((SmartResourceAgent)myAgent).getLocalName());
                                obj=obj.replaceAll("%Today%",today);

				if(pred_ns.equals(SmartResourceAgent.RDF_NS) && pred.equals("type")){
					if(obj_ns.equals(SmartResourceAgent.APL_NS)){
					   if(obj.equals("Behavior")){
						if(!((SmartResourceAgent)myAgent).activities.contains(new BehaviorRule(subj_uri))){
							activities_uris.add(subj_uri);
							BehaviorRule rule=new BehaviorRule(subj_uri);
							((SmartResourceAgent)myAgent).activities.add(rule);
						}
						else{
							print("Defining twice "+subj_uri+"!");
						}
					   }
					   else if(obj.equals("Belief")){
						belief_uris.add(subj_uri);
					   }
					   else if(obj.equals("Goal")){
						goal_uris.add(subj_uri);
					   }
					}
				}//type
				
				else{
				        if(belief_uris.contains(subj_uri)){
					   if(pred_ns.equals(SmartResourceAgent.APL_NS) && pred.equals("statement"))
					   	((SmartResourceAgent)myAgent).beliefs.add(new SemanticStatement(obj));
					   else print("Unknown predicate "+pred_ns+pred+" attached to "+subj_uri);
					}

					else if(goal_uris.contains(subj_uri)){
					   if(pred_ns.equals(SmartResourceAgent.APL_NS) && pred.equals("statement")) ((SmartResourceAgent)myAgent).goals.add(new SemanticStatement(obj));
					   else print("Unknown predicate "+pred_ns+pred+" attached to "+subj_uri);
					}

					else if(activities_uris.contains(subj_uri)){
						int ind=((SmartResourceAgent)myAgent).activities.lastIndexOf(new BehaviorRule(subj_uri));
						if(ind!=-1){
							BehaviorRule rule=(BehaviorRule)((SmartResourceAgent)myAgent).activities.elementAt(ind);
							if(pred_ns.equals(SmartResourceAgent.APL_NS)){
								if(pred.equals("class")) rule.classFile=obj;
								else if(pred.equals("inactive")) rule.inactive=obj;
								else if(pred.equals("trueIf")) rule.startInContext.addElement(new SemanticStatement(obj));
								else if(pred.equals("falseIf")) rule.doNotStartInContext.addElement(new SemanticStatement(obj));
								else if(pred.equals("addOnStart")) rule.addStatements.addElement(new SemanticStatement(obj));
								else if(pred.equals("removeOnStart")) rule.removeStatements.addElement(new SemanticStatement(obj));
								else if(pred.equals("trueIfGoalAchieved")) rule.needs.addElement(new SemanticStatement(obj));
								else if(pred.equals("achievesGoal")) rule.achieves.addElement(new SemanticStatement(obj));
								else if(pred.equals("event")) rule.event=new SemanticStatement(obj);
								else if(pred.equals("addOnSuccess"));
								else if(pred.equals("removeOnSuccess"));
								else if(pred.equals("addOnFail"));
								else if(pred.equals("removeOnFail"));
								else if(pred.equals("addOnEnd"));
								else if(pred.equals("removeOnEnd"));
								else print("Unknown predicate "+pred_ns+pred+" attached to "+subj_uri);
							}

							rule.startParams.add(new BehaviorStartParameter(pred_ns,pred,obj));
						}
					 
					}
					    
				}//not 'type'
			}
		
		}//handleStatement
   }


}
