package smartresource.core;

import smartresource.core.util.Logger;

import java.util.*;
import java.io.*;
import javax.swing.SwingUtilities;

import jade.core.Agent;
import jade.gui.GuiAgent;
import jade.gui.GuiEvent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;


public class SmartResourceAgent extends GuiAgent{

  public final static String RDF_NS="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
  public final static String APL_NS="http://www.smartresource.com/rgbdf#";
  public final static String RAB_NS="http://www.smartresource.com/atomic_behaviors#";
  public final static String SCRIPT_STATEMENT="I MissingScript ";
  public final static String CLASS_STATEMENT="I MissingClass ";
  
  private static boolean logCreated=false;
  
  public Vector beliefs; //current beliefs (and commitments), what is true now
  public Vector myRoles; //roles agent plays

  Vector activities; //behavioral rules
  Vector goals; //what is false now, but the agents wants to be true, for starting behaviours achieving those
  Vector goals_under_work; //in order not to start twice

  Live live; //main behavior, the deliberation cycle
  ServerEventsHandler serverHandler; //behavior handling events from HTTP

  Vector nonAvailableBehaviors; //behavior .class files that are missing

  SmartResourceAgentGUI myGUI; //agent's user interface, if needed
  SmartResourceAgentServer myServer; //agent's HTTP server (e.g. for getting events from external applications like Google Earth), if needed

  private Vector serverEventQueue;
  private Boolean serverEventQueueLock;
  Vector currentEvents; // used only in handleInterfaceEvent


//handle events from the user interface, defined in JADE
  protected void onGuiEvent(GuiEvent event){
  	String str=(String)event.getParameter(0);
	handleInterfaceEvent(str);
  }

//receives events from the HTTP server and puts to the queue
//to be later handled in agent's own thread
  public void postServerEvent(ServerEvent ev){
        String str=ev.message.replaceAll("%20"," ");
	if(str.length()>50) print("Received by http: "+str.substring(0,50)+"...("+str.length()+" chars)");
	else print("Received by http: "+str);

	synchronized(serverEventQueueLock){
		serverEventQueue.add(ev);
		serverHandler.restart();
	    }
  }

//handle events from HTTP server
  public boolean onServerEvent(ServerEvent ev){

        String str=ev.message.replaceAll("%20"," ");

	String response=handleInterfaceEvent(str);
	if(response==null) return false;

	if(response.equals("")) 
		response="HTTP/1.0 200 OK \r\nContent-Type: text/html\r\n\r\n <html><script type=\"text/javascript\">window.opener=self;window.close();</script></body></html>";

	try{
		OutputStream out=ev.socket.getOutputStream();
		PrintWriter pw=new PrintWriter(new OutputStreamWriter(out));
		pw.print(response);
		pw.flush();
		pw.close();
		ev.socket.close();
	}
	catch(IOException e){e.printStackTrace();}

	return true;
  }


//for cases when handling in the server thread is unavoidable
  public String processServerMessage(String message){
	doSuspend();

        String str=message.replaceAll("%20"," ");
	if(str.length()>50) print("Received by http: "+str.substring(0,50)+"...("+str.length()+" chars)");
	else print("Received by http: "+str);

	String response=handleInterfaceEvent(str);
	if(response==null || response.equals("")) 
		response="HTTP/1.0 200 OK \r\nContent-Type: text/html\r\n\r\n <html><script type=\"text/javascript\">window.opener=self;window.close();</script></body></html>";

	doActivate();

	return response;
  }


//actual handling of event from both interface and server
  private String handleInterfaceEvent(String str){

  	SemanticStatement s_ev=new SemanticStatement(str);
	currentEvents.add(s_ev);

        live.event=str;
	live.action();

	currentEvents.clear();
	restartLive();

	if(live.eventHandled) return live.response;
	else return null;
  }


  public void restartLive(){
	live.restart();
  }

  public void print(String str){
	Calendar now=Calendar.getInstance();
	String time=String.valueOf(now.get(Calendar.HOUR_OF_DAY))+":"+String.valueOf(now.get(Calendar.MINUTE))+":"+String.valueOf(now.get(Calendar.SECOND));
  	System.out.println("["+time+"] "+getLocalName()+": "+str);
  }

  protected void setup() {
	
	PrintStream oldOut=System.out;
	if(logCreated){}
	else{
	        logCreated=true;
		FileOutputStream fout;
		try{
			fout=new FileOutputStream("lastLog.txt");
			Logger log = new Logger(fout);
			log.oldOut=oldOut;
			System.setOut((PrintStream)log); 
			System.setErr((PrintStream)log); 
		}
		catch(Exception e){e.printStackTrace();}
	}
	
	serverEventQueue = new Vector();
	serverEventQueueLock = new Boolean(true);

	beliefs=new Vector();
	activities=new Vector();

	goals=new Vector();
	goals_under_work=new Vector();

	currentEvents=new Vector();

	nonAvailableBehaviors=new Vector();

	Object[] args = getArguments();
	String script_files="";

        myRoles=new Vector();

	script_files=args[0].toString();

	if(script_files.contains("startup.rdf") && args.length>1){
		String roles=args[1].toString();

		StringTokenizer token=new StringTokenizer(roles,"+");
		while(token.hasMoreTokens()){
			String role=token.nextToken();
			SemanticStatement ss=new SemanticStatement(SCRIPT_STATEMENT+role);
			beliefs.add(ss);
		}
	}
	
	StringTokenizer token=new StringTokenizer(script_files,"*");
	while(token.hasMoreTokens()){
		String script=token.nextToken();
		ParametrizedAsset ar=new AssignRoleBehavior();
		Vector params=new Vector();
		params.add(new BehaviorStartParameter(RAB_NS,"input",script));
		if(script.endsWith("startup.rdf"))
			params.add(new BehaviorStartParameter(RAB_NS,"startup","true"));		
		ar.setStartParameters(params);
		addBehaviour((Behaviour)ar);
	}

	addBehaviour(live=new Live());
	addBehaviour(serverHandler=new ServerEventsHandler());
  } 

  protected void takeDown() {
	print("NOOOOOO!!!!!!!");
	if(myServer!=null) myServer.stop();
  } 

  private class Live extends CyclicBehaviour{

 	Vector newGoals; // for whole action() iteration scope		
	Vector handledGoals; // for one rule scope

	String event;
	String response;
	boolean eventHandled;

	//evaluates the left part of the rule
	//returns the first fitting binding of variables
	//or NULL if the rule is not to be executed
	public HashMap evaluateRule(BehaviorRule b_rule){
			if(b_rule.inactive.equals("true")) return null; 

			boolean toBeStarted=true;

			VariablesBindingManager manager=new VariablesBindingManager();

			//event
			if(b_rule.event!=null){
				boolean found=manager.evaluate(b_rule.event, currentEvents,VariablesBindingManager.TRUE_IF);
				if(found){
					manager.adoptNewBindings();
				}
				else{ toBeStarted=false;}
			}

			//trueIf
			if(toBeStarted){
			   for(int i=0;i<b_rule.startInContext.size();i++){
				SemanticStatement st=new SemanticStatement(b_rule.startInContext.elementAt(i));
				
				boolean found=manager.evaluate(st, beliefs,VariablesBindingManager.TRUE_IF);

				if(found){
					manager.adoptNewBindings();
				}
				else{ toBeStarted=false; break;}
			   }
			}

			//need for this action?

			if(toBeStarted){
			   boolean needed=true;
			   if(b_rule.achieves.size()>0){
			   	needed=false;
			   	for(int i=0;i<b_rule.achieves.size();i++){
					SemanticStatement st=new SemanticStatement(b_rule.achieves.elementAt(i));

					boolean found=manager.evaluate(st,goals,VariablesBindingManager.TRUE_IF);
					if(found){
						manager.adoptNewBindings();

						HashMap vars=manager.getBinding(0);
						st.bindVars(vars);
						int ind=((SmartResourceAgent)myAgent).goals.indexOf(st);
						handledGoals.add((SemanticStatement)((SmartResourceAgent)myAgent).goals.elementAt(ind));
						needed=true;
					}
			   	}
			   }
			   if(!needed) toBeStarted=false;

			}

			//required resources available, subgoals achieved?
		        Vector subgoals=new Vector();
			if(toBeStarted){
			   for(int i=0;i<b_rule.needs.size();i++){
				SemanticStatement st=new SemanticStatement(b_rule.needs.elementAt(i));

				boolean found=manager.evaluate(st,beliefs,VariablesBindingManager.TRUE_IF);
				if(found){
					manager.adoptNewBindings();
				}
                                else{
					toBeStarted=false;
					HashMap vars=manager.getBinding(0);
					st.bindVars(vars);
					if(!goals.contains(st) && !goals_under_work.contains(st)){
						subgoals.add(st);
					}
				}
			   }

			}

			//falseIf
			for(int i=0;i<b_rule.doNotStartInContext.size();i++){
				SemanticStatement st=new SemanticStatement(b_rule.doNotStartInContext.elementAt(i));

				boolean found=manager.evaluate(st,beliefs,VariablesBindingManager.FALSE_IF);
				if(found){
        				toBeStarted=false;
					subgoals.clear();
					break;
				}
				else{
					manager.adoptNewBindings();
				}
			}

			newGoals.addAll(subgoals);

			if(toBeStarted) return manager.getBinding(0);
			else return null;
	}

	//generates startParameters where variables are replaced with values 
	public Vector bindVarsInParameters(Vector startParams, HashMap vars){
		Vector params=new Vector();
		for(int i=0;i<startParams.size();i++){
			BehaviorStartParameter bsp=(BehaviorStartParameter)startParams.elementAt(i);
			String newvalue=bsp.value;
					
			Iterator itvars=vars.entrySet().iterator();
			while(itvars.hasNext()){
				Map.Entry ent=(Map.Entry)itvars.next();
				String regex="\\*"+(String)ent.getKey()+"\\*";
				newvalue=newvalue.replaceAll(regex,(String)ent.getValue());
			}
			params.add(new BehaviorStartParameter(bsp.name_ns,bsp.name,newvalue));
		}
		return params;
	}

	public void action(){

	        goals.removeAll(beliefs);
	        goals_under_work.removeAll(beliefs);
	        
	        response=""; //for results from event handlers
	        eventHandled=true;

		Vector toAdd=new Vector(); //statements
		Vector toRemove=new Vector(); //statements

		newGoals=new Vector(); //statements		

		Iterator it=activities.iterator();
		while(it.hasNext()){

			handledGoals=new Vector(); //statements		
			
			BehaviorRule b_rule=(BehaviorRule)it.next();
			String b_class=b_rule.classFile;

			HashMap vars=evaluateRule(b_rule); //note it can also add to newGoals and handledGoals

			if(vars!=null){                     		
				//execute behavior
				try{
				    if(b_class!=null){
					Vector params=bindVarsInParameters(b_rule.startParams, vars);

					if(!b_class.startsWith("@")){
						Class cl=Class.forName(b_class);
						Object b=cl.newInstance();

						nonAvailableBehaviors.remove(b_class);

						goals.removeAll(handledGoals);
						goals_under_work.addAll(handledGoals);

						if(b instanceof InterfaceEventHandler){
							InterfaceEventHandler h=(InterfaceEventHandler)b;
							h.setAgent((SmartResourceAgent)myAgent);
							h.setStartParameters(params);
							h.onStart();
							response=response+h.handle(event);
							h.onEnd();
						}
						
						else if(b instanceof SmartResourceAgentGUI){
       						        SmartResourceAgentGUI gui=(SmartResourceAgentGUI)b;
							gui.setAgent((SmartResourceAgent)myAgent);
							((SmartResourceAgent)myAgent).myGUI=gui;

							gui.setStartParameters(params);
							gui.construct();
							gui.start();
						}

						else if(b instanceof SmartResourceAgentServer){
						        SmartResourceAgentServer server=(SmartResourceAgentServer)b;

							server.setAgent((SmartResourceAgent)myAgent);
							((SmartResourceAgent)myAgent).myServer=server;

							server.setStartParameters(params);
							server.start();
						}

						else if(b instanceof Behaviour){
							if(b instanceof ParametrizedAsset){
								((ParametrizedAsset)b).setStartParameters(params);
							}
	
							addBehaviour((Behaviour)b);
						}
					}
					else{
						goals.removeAll(handledGoals);
						goals_under_work.addAll(handledGoals);

						if(b_class.equals("@Print")){
						        String text=null;
							Iterator itt=params.iterator(); 
							while(itt.hasNext()){ 
								BehaviorStartParameter p=(BehaviorStartParameter)itt.next();
								if(p.name.equals("print")) text=p.value;
							}
							if(text!=null) print(text);
						}
						else if(b_class.equals("@ModifyGUI")){
							if(myGUI!=null){
									myGUI.setStartParameters(params);
									Runnable addIt = new Runnable(){
									public void run(){
										myGUI.modify();
									}
								};SwingUtilities.invokeLater(addIt);
							}
						}
					}
				    }	

				    for(int i=0;i<b_rule.addStatements.size();i++){
					SemanticStatement s=new SemanticStatement(b_rule.addStatements.elementAt(i));
					s.bindVars(vars);
				        toAdd.add(s);
				    }
				    for(int i=0;i<b_rule.removeStatements.size();i++){
					SemanticStatement s=new SemanticStatement(b_rule.removeStatements.elementAt(i));
					s.bindVars(vars);
					toRemove.add(s);
				    }

				}
				catch(ClassNotFoundException ce){
				        eventHandled=false;
					if(!nonAvailableBehaviors.contains(b_class)){
						print("atomic behavior "+b_class+" is missing");
						nonAvailableBehaviors.add(b_class);
						SemanticStatement ss=new SemanticStatement(CLASS_STATEMENT+b_class.replace('.','/'));
						toAdd.add(ss);
					}
				}
				catch(Exception e){e.printStackTrace();}

			}

		}//rules iterator

		for(int i=0;i<toRemove.size();i++){
			SemanticStatement st=(SemanticStatement)toRemove.elementAt(i);
		        boolean found;
		        do{
				found=beliefs.remove(st);
			}while(found);
		}

		for(int i=0;i<toAdd.size();i++){
			SemanticStatement st=(SemanticStatement)toAdd.elementAt(i);
			beliefs.add(st);
		}

		for(int i=0;i<newGoals.size();i++){
			SemanticStatement st=(SemanticStatement)newGoals.elementAt(i);
			goals.add(st);
		}


//		if(getLocalName().equals("operator"))
//		{String bbb=""; Iterator it1=beliefs.iterator(); while(it1.hasNext()) bbb=bbb+((SemanticStatement)it1.next()).toString()+"; "; print(bbb);}
//		if(getLocalName().equals("test"))
//		{String ccc=" ";Iterator it1=goals.iterator(); while(it1.hasNext()) ccc=ccc+((SemanticStatement)it1.next()).toString()+"; "; print(ccc);}
//		if(getLocalName().equals("test"))
//		{String bbb=""; Iterator it1=activities.iterator(); while(it1.hasNext()) bbb=bbb+((BehaviorRule)it1.next()).uri+"; "; print(bbb);}


		if(toAdd.size()==0 && toRemove.size()==0 && newGoals.size()==0){
			block();
		}

	}

  }//Live behavior

  private class ServerEventsHandler extends CyclicBehaviour{

	public void action(){
		if (!serverEventQueue.isEmpty()){
			ServerEvent ev=null;  				
			synchronized(serverEventQueueLock){
				ev =(ServerEvent)serverEventQueue.elementAt(0);
				serverEventQueue.removeElementAt(0);
			}
			boolean result=onServerEvent(ev);
			if(!result){
				synchronized(serverEventQueueLock){
					serverEventQueue.add(ev);
				}
			}
		}
		else block();
	}
  }


}
